/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "mgmt.h"
#include "iov_str.h"
#include "digmd5.h"

static char *
get_path(struct iovec *dst, struct mog_mgmt *mgmt, char *buf, bool sdup)
{
	dst->iov_base = buf + mgmt->mark[0];
	dst->iov_len = mgmt->mark[1] - mgmt->mark[0];

	if (mog_valid_path(dst->iov_base, dst->iov_len)) {
		char *path;

		if (sdup) {
			path = xmalloc(dst->iov_len + 1);
			memcpy(path, dst->iov_base, dst->iov_len);
		} else {
			path = dst->iov_base;
		}

		path[dst->iov_len] = '\0';
		return path;
	} else {
		struct iovec iov;

		IOV_STR(&iov, "ERROR: uri invalid (contains ..)\r\n");
		mog_mgmt_writev(mgmt, &iov, 1);
		return NULL;
	}
}

/* starts the MD5 request */
void mog_mgmt_fn_md5(struct mog_mgmt *mgmt, char *buf)
{
	struct iovec iov[2];
	char *path = get_path(iov, mgmt, buf, true);

	if (!path) return;

	mgmt->forward = mog_file_open_read(mgmt->svc, path);
	if (mgmt->forward) {
		struct mog_file *file = &mgmt->forward->as.file;

		mog_fadv_noreuse(mgmt->forward->fd, 0, 0 /* ALL */);
		mog_fadv_sequential(mgmt->forward->fd, 0, 0 /* ALL */);
		file->md5 = mog_md5_new();
		file->pathlen = iov[0].iov_len;
		file->path = path;
	} else {
		IOV_STR(&iov[1], " MD5=-1\r\n");
		mog_mgmt_writev(mgmt, iov, 2);
		free(path);
	}
}

/* finishes the MD5 request */
#define CLEN(s) (sizeof(s)-1)
#define PFX " MD5="

void mog_mgmt_fn_md5_err(struct mog_mgmt *mgmt)
{
	struct iovec iov[3];
	struct mog_fd *mfd = mgmt->forward;
	struct mog_file *file = &mfd->as.file;
	long long offset = (long long)lseek(mfd->fd, 0, SEEK_CUR);
	char buf[sizeof(" at 18446744073709551615 failed\r\n") - 1];

	/* offset could be -1 here, but there ain't much we can do */

	IOV_STR(iov, "ERR read ");
	iov[1].iov_base = file->path;
	iov[1].iov_len = file->pathlen;
	iov[2].iov_base = buf;
	iov[2].iov_len = snprintf(buf, sizeof(buf),
	                          " at %lld failed\r\n", offset);
	mog_mgmt_writev(mgmt, iov, 3);
}

/* output: "/$PATH MD5=hex\r\n" */
void mog_mgmt_fn_md5_emit(struct mog_mgmt *mgmt)
{
	struct iovec iov[2];
	char buf[CLEN(PFX) + 32 + CLEN("\r\n")];
	char *b;
	struct mog_fd *mfd = mgmt->forward;
	struct mog_file *file = &mfd->as.file;

	iov[0].iov_base = file->path;
	iov[0].iov_len = file->pathlen;

	iov[1].iov_base = buf;
	iov[1].iov_len = sizeof(buf);

	b = mempcpy(buf, PFX, CLEN(PFX));
	mog_md5_hex(file->md5, b, 32);
	b[32] = '\r';
	b[33] = '\n';
	mog_mgmt_writev(mgmt, iov, 2);
}

/*
 * writes to mgmt fd:
 *   "URI $SIZE\r\n" on success
 *   "URI -1\r\n" on failure
 *   "ERROR: uri invalid (contains ..)\r\n" on invalid paths
 */
void mog_mgmt_fn_size(struct mog_mgmt *mgmt, char *buf)
{
	struct stat sb;
	struct iovec iov[2];
	char tmp[sizeof(" 18446744073709551615\r\n") - 1];
	char *path = get_path(iov, mgmt, buf, false);

	if (!path) return;

	if (mog_stat(mgmt->svc, path, &sb) == 0) {
		long long size = (long long)sb.st_size;

		iov[1].iov_base = tmp;
		iov[1].iov_len = snprintf(tmp, sizeof(tmp), " %lld\r\n", size);
	} else {
		IOV_STR(&iov[1], " -1\r\n");
	}

	mog_mgmt_writev(mgmt, iov, 2);
}

void mog_mgmt_fn_blank(struct mog_mgmt *mgmt)
{
	struct iovec iov;

	IOV_STR(&iov, "\r\n");
	mog_mgmt_writev(mgmt, &iov, 1);
}

void mog_mgmt_fn_unknown(struct mog_mgmt *mgmt, char *buf)
{
	struct iovec iov[3];

	IOV_STR(&iov[0], "ERROR: unknown command: ");
	iov[1].iov_base = mgmt->mark[0] + buf;
	iov[1].iov_len = mgmt->mark[1] - mgmt->mark[0];
	IOV_STR(&iov[2], "\r\n");
	mog_mgmt_writev(mgmt, iov, 3);
}

void mog_mgmt_fn_watch_err(struct mog_mgmt *mgmt)
{
	struct iovec iov;

	IOV_STR(&iov, "ERR iostat unavailable\r\n");
	mog_mgmt_writev(mgmt, &iov, 1);
}
