/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

/*
 * we want activeq_lock to be on a different cache line than
 * activeq_head and active_fd so we don't end up triggering
 * synchronization if a lock is contended.  We try to ensure
 * 64-bit(+) CPUs get past a (common) cache line size of 64
 * and 32-bit CPUs have a cache line size of 32.
 *
 * 128 byte cache lines seem to be getting rarer and (probably)
 * aren't worth optimizing for. (struct mog_fd is 128-bytes itself)
 */
verify(offsetof(struct mog_fd, as.queue.activeq_lock)
       >= (sizeof(long) >= 8 ? 64 : 32));

/*
 * access to this should only be called in the main thread, this
 * is currently not thread safe as there's no need for it.
 */
static LIST_HEAD(queue_head, mog_queue) all_queues;

/*
 * used only by queue_epoll.c, but should be usable with queue_kqueue.c
 * in the future
 */
struct mog_queue *mog_queue_init(int queue_fd)
{
	struct mog_fd *mfd;
	struct mog_queue *q;

	CHECK(int, 0, mog_set_cloexec(queue_fd, true));

	mfd = mog_fd_init(queue_fd, MOG_FD_TYPE_QUEUE);
	q = &mfd->as.queue;
	q->queue_fd = queue_fd;
	SIMPLEQ_INIT(&q->activeq_head);
	CHECK(int, 0, pthread_mutex_init(&q->activeq_lock, NULL));
	memset(&q->thrpool, 0, sizeof(struct mog_thrpool));
	LIST_INSERT_HEAD(&all_queues, q, qbuddies);

	return q;
}

void mog_queue_stop(struct mog_queue *keep)
{
	struct mog_queue *queue, *tmp;
	struct mog_fd *mfd;

	LIST_FOREACH_SAFE(queue, &all_queues, qbuddies, tmp) {
		/* keep is usually mog_notify_queue */
		if (queue == keep)
			continue;
		LIST_REMOVE(queue, qbuddies);
		mog_thrpool_quit(&queue->thrpool);
		CHECK(int, 0, pthread_mutex_destroy(&queue->activeq_lock));
		mfd = mog_fd_of(queue);
		mog_fd_put(mfd);
	}
}
